# {{ crate_name }}

This Rust crate was generated by `gradgen`.

Backend mode: `{{ backend_mode }}`
Scalar type: `{{ scalar_type }}`
{% if math_library %}
Math library namespace: `{{ math_library }}`
{% endif %}

## Build

```bash
cargo build
```

## Test Integration

Add this crate to your Rust project or call the generated function directly from `src/lib.rs`.

## Generated Function

- Function name: `{{ codegen.function_name }}`
- Workspace size: `{{ codegen.workspace_size }}`
- Input sizes: `{{ codegen.input_sizes }}`
- Output sizes: `{{ codegen.output_sizes }}`

The generated ABI uses input slices, output slices, and a mutable workspace slice.
The generated Rust file also includes helper functions that return the required
workspace, input, and output dimensions.
The crate root also contains a `metadata.json` file with the same function
metadata in a machine-readable form.
{% if backend_mode == "no_std" %}
The generated crate is `no_std` and uses `{{ math_library }}` for transcendental math functions.
{% endif %}
